from django.db import models
from django.utils.translation import ugettext_lazy as _

from isbn_field import ISBNField


class Book(models.Model):
    """A book that is available for borrowing by persons."""

    title = models.CharField(verbose_name=_("Book title"), max_length=50)
    author = models.CharField(verbose_name=_("Author name"), max_length=50, blank=True)
    publisher = models.CharField(verbose_name=_("Publishing company"), max_length=50, blank=True)

    edition = models.PositiveSmallIntegerField(
        verbose_name=_("Number of edition"), blank=True, null=True
    )
    year = models.PositiveSmallIntegerField(
        verbose_name=_("Publishing year"), blank=True, null=True
    )

    isbn = ISBNField(verbose_name=_("ISBN number"), unique=True)

    cover = models.ImageField(verbose_name=_("Photo of cover"), blank=True, null=True)

    def __str__(self) -> str:
        return "%s (%s, %d. %d %d)" % (
            self.title,
            self.author,
            self.edition,
            _("edition"),
            self.year,
        )

    def is_deleteable(self) -> bool:
        for book_copy in self.copies.all():
            if not book_copy.is_deleteable():
                return False

        return True

    class Meta:
        unique_together = [["title", "author", "edition", "year"]]
        ordering = ["title", "year", "edition"]


class BookCopy(models.Model):
    """A single copy of a book. This is the item that will be associated with
    a person who borrowed it.
    """

    CONDITION_CHOICES = [
        ("U", _("New, unused")),
        ("N", _("New, used")),
        ("S", _("Slightly worn")),
        ("H", _("Heavily worn")),
    ]

    book = models.ForeignKey(
        "Book", blank=True, null=True, on_delete=models.CASCADE, related_name="copies"
    )
    borrower = models.ForeignKey(
        "core.Person", blank=True, null=True, on_delete=models.CASCADE, related_name="books",
    )

    copy_num = models.PositiveSmallIntegerField(
        verbose_name=_("Current number of copy"), blank=True, null=True
    )

    borrow_count = models.PositiveSmallIntegerField(
        verbose_name=_("Number of borrowings"), blank=True, null=True
    )
    condition = models.CharField(
        verbose_name=_("Condition"), choices=CONDITION_CHOICES, max_length=1, blank=True
    )
    remarks = models.CharField(verbose_name=_("Remarks"), max_length=100, blank=True)

    barcode = models.CharField(verbose_name=_("Barcode"), max_length=20, unique=True)

    def save(self, *args, **kwargs) -> None:
        if not self.copy_num:
            if self.book.copies.exclude(copy_num__isnull=True).exists():
                self.copy_num = (
                    self.book.copies.exclude(copy_num__isnull=True)
                    .order_by("copy_num")
                    .last()
                    .copy_num
                    + 1
                )
            else:
                self.copy_num = 1

        if not self.barcode:
            self.barcode = "%s-%s" % (self.book.isbn, self.copy_num)

        super().save(*args, **kwargs)

    def is_deleteable(self) -> bool:
        return not self.borrow_count and not self.borrower

    class Meta:
        ordering = ["book__title", "book__year", "book__edition", "copy_num"]
        indexes = [models.Index(fields=["barcode"])]
